package se.chalmers.tdv.ddmin;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Delta Debugging. Minimising input.
 * 
 * Currently relies on the test method to not change the values of the inputs
 * (circumstances), since those changes would also change the input under test.
 * If changes to the circumstances are necessary, the test method needs to
 * perform a deep copy first.
 * 
 * 
 * @param <T>
 *            Type of the circumstances under test.
 */
public class DDMin<T> {

	/**
	 * The test function, returning one of PASS, FAIL or UNRESOLVED.
	 */
	private DDTest<T> ddTest;

	// print debug messages.
	public static boolean DEBUG = false;

	/**
	 * Test case succeeded.
	 */
	public static final int PASS = 0;

	/**
	 * Test case failed.
	 */
	public static final int FAIL = 1;

	/**
	 * Test produced indeterminate results.
	 */
	public static final int UNRESOLVED = 2;

	/**
	 * Construct a new DDMin instance.
	 * 
	 * @param ddTest
	 *            Implementation of the test to be minimised against.
	 */
	public DDMin(DDTest<T> ddTest) {
		this.ddTest = ddTest;
	}

	/**
	 * 
	 * @param circumstances
	 *            <p>
	 *            A failing test case, split into the 'circumstances' that make
	 *            up this test case. Examples:
	 *            </p>
	 *            <ul>
	 *            <li>if the test case is an array, the circumstances can simply
	 *            be the elements in that array.
	 *            <li>if the test case is a series of user-triggered events, the
	 *            circumstances can be the individual events.
	 *            <li>if the test case is a program, the circumstances can be
	 *            the individual statements.
	 *            </ul>
	 *            <p>
	 *            When test cases are more structured than just the union of
	 *            their elements (e.g. a program) it is suggested to implement
	 *            the method {@link DDTest#validInput(List) validInput} in
	 *            DDTest to exclude circumstances that don't form a valid input.
	 *            </p>
	 * @return
	 */
	public Collection<T> ddmin(Collection<T> circumstances) {
		if (this.ddTest.test(circumstances) != FAIL)
			throw new IllegalArgumentException(
					"Test should fail on provided circumstances.");
		if (this.ddTest.test(new LinkedList<T>()) == FAIL) {
			return new LinkedList<T>();
		}
		return ddmin(circumstances, 2);
	}

	private Collection<T> ddmin(Collection<T> circumstances, int nChunks) {

		if (DEBUG)
			System.out.println("ddmin " + nChunks + " - " + circumstances);

		// Case 1: cannot divide into chunks.
		if (circumstances.size() == 0)
			return circumstances;

		Collection<Collection<T>> chunks = splitInChunks(circumstances, nChunks);

		// Case 2: one chunk can be removed, input still fails.
		for (Collection<T> c : chunks) {
			Collection<T> smallerInput = allChunksExcept(chunks, c);
			if (this.ddTest.test(smallerInput) == FAIL)
				return ddmin(smallerInput, Math.max(nChunks - 1, 2));
		}

		// Case 3: no chunk can be removed, try to increase number of chunks,
		// making each chunk smaller.
		if (nChunks < circumstances.size())
			return ddmin(circumstances,
					Math.min(2 * nChunks, circumstances.size()));

		// Case 4: no chunk can be removed, size of chunk cannot be reduced
		// further. The input is minimal.
		return circumstances;

	}

	// Split the provided collection into n chunks. Last chunk may be shorter
	// than others.
	private Collection<Collection<T>> splitInChunks(
			Collection<T> circumstances, int n) {

		// Any collection should work; using lists instead of sets to allow
		// duplicates.
		Collection<T> chunk = new LinkedList<T>();
		Collection<Collection<T>> result = new LinkedList<Collection<T>>();

		int chunkMaxSize = circumstances.size() / n;

		// create chunks of the right size, add them to the result.
		for (T c : circumstances) {
			chunk.add(c);
			if (chunk.size() == chunkMaxSize) {
				result.add(chunk);
				chunk = new LinkedList<T>();
			}
		}
		if (chunk.size() > 0)
			result.add(chunk);

		return result;
	}

	// Combine all chunks into one collection, leaving the specified collection
	// out.
	private Collection<T> allChunksExcept(Collection<Collection<T>> chunks,
			Collection<T> excluded) {
		Collection<T> result = new LinkedList<T>();

		for (Collection<T> c : chunks)
			if (c != excluded)
				result.addAll(c);

		return result;
	}
}
