package se.chalmers.tdv.testing;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Random;

import se.chalmers.tdv.ddmin.DDMin;
import se.chalmers.tdv.ddmin.DDTest;

/**
 * A simple implementation of the ideas described in the paper
 * "Testing Monadic Code with QuickCheck" by Koen Claessen and John Hughes,
 * Haskell Workshop 2002. Available at
 * http://www.cse.chalmers.se/~rjmh/Papers/QuickCheckST.ps
 * 
 * Implementations of this generic class allow for the testing of observational
 * equivalence of actions on stateful objects. The action sequences under test
 * can be surrounded by a common random prefix and suffix of actions. It is
 * possible to specify a precondition on the test before it is run.
 * 
 * Implementing tests can test for additional functional properties by
 * overwriting the test function inherited from StatefulTest.
 * 
 * There is some support for minimization, using the ddmin algorithm on the
 * prefix and suffix that form the context.
 * 
 * @param <T>
 *            The type of the object under test.
 * @param <Action>
 *            A data structure representing the different actions that can be
 *            performed on the type of object under test.
 * @param <Out>
 *            Type of potential output values.
 */
public abstract class StatefulTestCase<T, Action, Out> {

	/** Limiting the length of prefix and suffix. */
	public static final int MAX_PREFIX_LENGTH = 10;
	public static final int MAX_SUFFIX_LENGTH = 10;

	/** The commands to test for observational equivalence. */
	protected Action[] command1;
	protected Action[] command2;

	/** Two objects used to test those commands. */
	protected T object1;
	protected T object2;

	/** The context surrounding the command under test. */
	protected Action[] commonPrefix;
	protected Action[] commonSuffix;

	/** The test of which this test case is an instance. */
	protected StatefulTest<T, Action, Out> currentTest;

	/** Sequence of outputs. */
	public LinkedList<Out> outputArray1;
	public LinkedList<Out> outputArray2;
	private boolean printException;

	public StatefulTestCase(StatefulTest<T, Action, Out> test) {
		this.currentTest = test;
	}

	public T getObject1() {
		return this.object1;
	}

	public T getObject2() {
		return this.object2;
	}

	public void setCommands(Action[] com1, Action[] com2) {
		this.command1 = com1;
		this.command2 = com2;
	}

	public void setObjects(T obj1, T obj2) {
		this.object1 = obj1;
		this.object2 = obj2;
	}

	/**
	 * Test the observable equivalence of the two commands.
	 * 
	 * @return True if, after performing the prefix, command and suffix on each
	 *         respective object the object's are marked as equal by their
	 *         'equals' method. False if the objects are not equal, or if an
	 *         exception occurred during the execution of the actions.
	 */
	private boolean testEquivalence() {
		outputArray1 = new LinkedList<>();
		outputArray2 = new LinkedList<>();
		if (currentTest.withPrefix())
			for (Action action : commonPrefix) {
				perform(object1, action, outputArray1);
				perform(object2, action, outputArray2);
			}
		for (Action action : command1) {
			perform(object1, action, outputArray1);
		}
		for (Action action : command2) {
			perform(object2, action, outputArray2);
		}
		if (currentTest.withSuffix())
			for (Action action : commonSuffix) {
				perform(object1, action, outputArray1);
				perform(object2, action, outputArray2);
			}
		return object1.equals(object2);
	}

	public boolean testObservableEquivalence() {
		try {
			return testEquivalence();
		} catch (Exception e) {
			if (printException) {
				System.out.println();
				System.out.print("Exception when running test case: ");
				e.printStackTrace();
			}
			return false;
		}
	}

	/** Setting a random prefix and suffix of actions. */
	public void initializeContext() {
		Random rand = new Random();
		commonPrefix = randomActions(rand.nextInt(MAX_PREFIX_LENGTH + 1));
		commonSuffix = randomActions(rand.nextInt(MAX_SUFFIX_LENGTH + 1));
	}

	/**
	 * Initializes the test case until the precondition is met, and then run the
	 * test.
	 * 
	 * @return True if the test passed, false if failed.
	 */
	public boolean initializeAndRunTest() {
		do {
			this.currentTest.initialize(this);
			this.initialize();
			this.initializeContext();
		} while (!this.precondition());
		printException = true;
		boolean res = runTest();
		printException = false;
		return res;
	}

	public boolean runTest() {
		this.setUp();
		return this.currentTest.test(this);
	}

	/** Prints a description of this test case to the standard out. */
	public void print() {
		
		this.runTest();
		if (currentTest.withPrefix())
			System.out
					.println("  Prefix:     " + Arrays.toString(commonPrefix));
		if (currentTest.withSuffix())
			System.out
					.println("  Suffix:     " + Arrays.toString(commonSuffix));
		System.out.println("  Commands 1: " + Arrays.toString(command1));
		System.out.println("  Commands 2: " + Arrays.toString(command2));
		System.out.println("  Result 1:   " + object1.toString());
		System.out.println("  Result 2:   " + object2.toString());
		System.out.println("  Output 1:   " + outputArray1.toString());
		System.out.println("  Output 2:   " + outputArray2.toString());
	}

	/**
	 * Interleaving, minimizes the prefix and suffix until they are not further
	 * minimized. When returning, the objects are in the state after running the
	 * commands in minimized context.
	 */
	public void minimizeContext() {
		int prefixLength, suffixLength;
		do {
			prefixLength = commonPrefix.length;
			suffixLength = commonSuffix.length;
			commonPrefix = minimizeContext(true);
			commonSuffix = minimizeContext(false);

		} while (prefixLength > commonPrefix.length
				|| suffixLength > commonSuffix.length);

		this.setUp();
		testObservableEquivalence();

	}

	// Calls the dd-min algorithm on the prefix if the boolean set to true, on
	// the suffix otherwise. Returns the minimized sequence of actions.
	@SuppressWarnings("unchecked")
	private Action[] minimizeContext(final boolean minPrefix) {

		final StatefulTestCase<T, Action, Out> self = this;
		DDTest<Action> test = new DDTest<Action>() {
			public int test(Collection<Action> circumstances) {
				if (minPrefix)
					commonPrefix = (Action[]) circumstances.toArray();
				else
					commonSuffix = (Action[]) circumstances.toArray();
				if (minPrefix && !precondition())
					return DDMin.PASS;
				self.setUp();
				if (currentTest.test(self))
					return DDMin.PASS;
				return DDMin.FAIL;
			}
		};
		DDMin<Action> minimizer = new DDMin<Action>(test);
		return (Action[]) minimizer.ddmin(
				new ArrayList<Action>(Arrays.asList(minPrefix ? commonPrefix
						: commonSuffix))).toArray();
	}

	/** Perform the specified action on the provided object. */
	public abstract void perform(T object, Action action, LinkedList<Out> out);

	/** Called once to initialize data for this test case. */
	public abstract void initialize();

	/**
	 * Called before each test to set up the context. This method should at
	 * least make a call to the setObjects function to initialize the objects
	 * under test.
	 */
	public abstract void setUp();

	/** Returns a random sequence of n actions. */
	public abstract Action[] randomActions(int n);

	/** Return true if this test case matches the precondition. */
	public boolean precondition() {
		return true;
	}

}
